package com.ef_prime.rflow.io;

import java.awt.Dimension;
import java.awt.Point;
import java.awt.geom.Point2D;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import com.ef_prime.rflow.FlowSetting;
import com.ef_prime.rflow.explorer.FunctionNode;
import com.ef_prime.rflow.explorer.PartsNode;
import com.ef_prime.rflow.flow.ViewLayout;
import com.ef_prime.rflow.flow.model.FlowEdge;
import com.ef_prime.rflow.flow.model.FlowGraph;
import com.ef_prime.rflow.flow.model.FlowNode;
import com.ef_prime.rflow.group.AppearanceGroup;
import com.ef_prime.rflow.group.AppearanceGroupManager;
import com.ef_prime.rflow.group.DefaultAppearanceGroup;
import com.ef_prime.rflow.group.IconAppearanceGroup;
import com.ef_prime.rflow.group.ShapeAppearanceGroup;
import com.ef_prime.rflow.node.NodeModel;
import com.ef_prime.rflow.node.NodeModelController;
import com.ef_prime.rflow.node.NodeModelManager;
import com.ef_prime.rflow.node.base.SubflowNodeModel;
import com.ef_prime.rflow.xml.XMLException;
import com.ef_prime.rflow.xml.XMLWriter;

import edu.uci.ics.jung.graph.Vertex;

public class ElementWriterFactory{
	
	public static ElementWriter getGraphWriter(ViewLayout layout, FlowSetting fileSettings){
		return new GraphWriter(layout, fileSettings); 
	}
	
	public static ElementWriter getFunctionTreeElementWriter(AppearanceGroupManager groupManager, FunctionNode functionRoot){
		return new FunctionTreeElementWriter(groupManager, functionRoot);
	}
	
	public static ElementWriter getPartsTreeWriter(PartsNode partsRoot){
		return new PartsTreeWriter(partsRoot);
	}
	
	private static abstract class AbstractGraphWriter implements ElementWriter, FlowDataElements{

		protected FlowGraph graph;
		
		AbstractGraphWriter(FlowGraph graph){
			this.graph = graph;
		}
		
		abstract Point getLayoutOffset();
		
		abstract Point2D getLocation(FlowNode node);
		
		abstract Dimension getSize();
		
		abstract String getLocationType();
		
		abstract void writeFileSettings(XMLWriter out) throws XMLException ;
		
		public void write(XMLWriter out) throws XMLException {
			if(graph == null)
				return;
			
			//node -> id MAP エッジ保存用
			Map<Vertex, Integer> nodeToIdMap = new HashMap<Vertex, Integer>();
			
			Dimension size = getSize();
			
			out.writeStartElement(GRAPH);
			out.writeAttribute(VERSION, GRAPH_FILE_VERSION);
			if(size != null){
				out.writeAttribute(WIDTH, String.valueOf(size.width));
				out.writeAttribute(HEIGHT, String.valueOf(size.height));
			}
			out.writeAttribute(LOCATION_TYPE, getLocationType());
			Point offset = getLayoutOffset();
			if(offset != null){
				out.writeAttribute(LOCATION_OFFSET_X, String.valueOf(offset.x));
				out.writeAttribute(LOCATION_OFFSET_Y, String.valueOf(offset.y));
			}
			//save layouttrans
			
			//ファイルごと設定
			writeFileSettings(out);
			
			for(Object ob_v : graph.getVertices()){
				FlowNode node = (FlowNode)ob_v;
				NodeModel nodeModel = node.getModel();
				
				int id = nodeToIdMap.size();
				nodeToIdMap.put(node, id);
				
				Point2D p = getLocation(node);
				
				out.writeStartElement(NODE);
				out.writeAttribute(ID, String.valueOf(id));
				if(nodeModel.isTitle())
					out.writeAttribute(TITLE, nodeModel.getTitle());
				/*相対位置
				out.writeAttribute(X, String.valueOf((float)p.getX()));
				out.writeAttribute(Y, String.valueOf((float)p.getY()));
				*/
				out.writeAttribute(X, String.valueOf((int)p.getX()));
				out.writeAttribute(Y, String.valueOf((int)p.getY()));
				
				
				out.writeStartElement(COMMAND);
				out.writeCharacters(nodeModel.getCommand());
				out.writeEndElement();
				
				//プロパティ
				out.writeStartElement(PROPERTY);
				if(nodeModel.getProperty().getAppearanceGroup() != null){
					out.writeAttribute(SHAPE, DefaultAppearanceGroup.valueOf(nodeModel.getProperty().getAppearanceGroup()).toString());
				}
				if(nodeModel.getProperty().getComment() != null){
					out.writeStartElement(COMMENT);
					out.writeCharacters(nodeModel.getProperty().getComment());
					out.writeEndElement();//comment
				}
				out.writeEndElement();//property
				
		
				out.writeStartElement(OPTION);
				out.writeAttribute(TYPE, nodeModel.getClass().getCanonicalName());
				
				
				if(nodeModel instanceof SubflowNodeModel){
					SubflowNodeModel boxNodeModel = (SubflowNodeModel)nodeModel;
					FlowDataWriter writer = new FlowDataWriter();
					writer.write(out, new ElementWriter[] { 
							getGraphWriter(boxNodeModel.getLayout(), null) },
							false, false);
				}else{
					NodeModelController controller = NodeModelManager.getModelController(nodeModel.getClass());
					Properties properties = null;
					if(controller != null)
						properties = controller.getOptionProperties(nodeModel);					
					
					if(properties != null){
						for(Map.Entry<Object, Object> entry : properties.entrySet()){
							out.writeStartElement(ENTRY);
							out.writeAttribute(KEY, (String)entry.getKey());
							out.writeCharacters((String)entry.getValue());
							out.writeEndElement();
						}
					}
					
				}
				out.writeEndElement();//option
				
				
				out.writeEndElement();//node
			}
			
			for(Object ob_e : graph.getEdges()){
				FlowEdge e = (FlowEdge)ob_e;
				out.writeEmptyElement(EDGE);
				out.writeAttribute(FROM, nodeToIdMap.get(e.getSource()).toString());
				out.writeAttribute(TO, nodeToIdMap.get(e.getDest()).toString());
			}
			
			out.writeEndElement();//graph
		}
	}
	
	private static class GraphWriter extends AbstractGraphWriter{

		private ViewLayout layout;
		private FlowSetting fileSettings;
		//private Point zeroPoint;
		
		@SuppressWarnings("unchecked") //$NON-NLS-1$
		public GraphWriter(ViewLayout layout, FlowSetting fileSettings){
			super(layout.getGraph());
			this.layout = layout;
			this.fileSettings = fileSettings;
			if(this.fileSettings != null)
				this.fileSettings.update();
		}
		
		@Override
		Point getLayoutOffset() {
			return layout.getOffset();
		}
		
		/*相対位置
		@Override
		Point2D getLocation(FlowNode node) {
			Point2D p = viewer.getGraphLayout().getLocation(node);
			p = viewer.transform(p);
			Dimension d = viewer.getSize();
			return new Point2D.Double(p.getX()/d.width, p.getY()/d.height);
		}
		*/
		
		/*絶対位置
		@Override
		Point2D getLocation(FlowNode node) {
			Point2D p = viewer.getGraphLayout().getLocation(node);
			p = viewer.layoutTransform(p);
			return new Point2D.Float((int)p.getX(), (int)p.getY());
		}*/
		
		@Override
		Point2D getLocation(FlowNode node) {
			return layout.getLocation(node);
		}
		
		@Override
		Dimension getSize() {
			return layout.getSize();
		}
		
		@Override
		String getLocationType() {
			return LOCATION_TYPE_ABSOLUTE;
		}
		
		@Override
		void writeFileSettings(XMLWriter out) throws XMLException {
			if(fileSettings == null)
				return ;
			
			out.writeStartElement(FLOWSETTING);

			for(Map.Entry<Object, Object> entry : fileSettings.entrySet()){
				out.writeStartElement(ENTRY);
				out.writeAttribute(KEY, (String)entry.getKey());
				out.writeCharacters((String)entry.getValue());
				out.writeEndElement();
			}
			
			out.writeEndElement();//FILESETTINGS
		}
	}
	
	private static class FunctionTreeElementWriter implements ElementWriter, FlowDataElements{

		private AppearanceGroupManager groupManager; 
		private FunctionNode functionRoot;
		
		public FunctionTreeElementWriter(AppearanceGroupManager groupManager, FunctionNode functionRoot) {
			this.functionRoot = functionRoot;
			this.groupManager = groupManager;
		}

		public void write(XMLWriter out) throws XMLException {
			if(functionRoot == null)return;
		
			out.writeStartElement(FUNCTION_TREE);
			out.writeAttribute(VERSION, FUNCTION_TREE_VERSION);
			writeFunctionGroup(out);
			writeFunctionTreeNode(out, functionRoot);
			out.writeEndElement();
		}
		
		private void writeFunctionGroup(XMLWriter out) throws XMLException{
			for(AppearanceGroup group : groupManager.getGroups()){
				out.writeEmptyElement(FUNCTION_GROUP);
				out.writeAttribute(TITLE, group.getTitle());
				if (group instanceof IconAppearanceGroup) {
					IconAppearanceGroup iconGroup = (IconAppearanceGroup) group;
					out.writeAttribute(ICON, iconGroup.getIconPath());
				}else if (group instanceof ShapeAppearanceGroup) {
					ShapeAppearanceGroup shapeGroup = (ShapeAppearanceGroup) group;
					out.writeAttribute(SHAPE, shapeGroup.getNodeShape().toString());
				}
			}
		}

		private void writeFunctionTreeNode(XMLWriter out, FunctionNode node) throws XMLException{
			out.writeStartElement(FUNCTION_NODE);
			out.writeAttribute(TITLE, node.getTitle());
			
			if(node.hasFunction()){
				out.writeAttribute(FUNCTION, node.getFunction());
			}
			if(node.hasGroup()){
				out.writeAttribute(GROUP, node.getGroup().getTitle());
			}
			if(node.hasSource()){
				out.writeStartElement(SOURCE);
				out.writeCharacters(node.getSource());
				out.writeEndElement();
			}
			
			if(!node.isLeaf()){
				for(int i = 0; i < node.getChildCount(); i++){
					FunctionNode child = (FunctionNode)node.getChildAt(i);
					writeFunctionTreeNode(out, child);
				}
			}
			
			out.writeEndElement();
		}
		
	}
	
	private static class PartsTreeWriter implements ElementWriter, FlowDataElements{
		
		private PartsNode partsRoot;
		
		public PartsTreeWriter(PartsNode partsRoot) {
			this.partsRoot = partsRoot;
		}
		
		public void write(XMLWriter out) throws XMLException {
			if(partsRoot == null)return;
		
			out.writeStartElement(PARTS_TREE);
			writePartsNode(out, partsRoot);
			out.writeEndElement();
		}

		private void writePartsNode(XMLWriter out, PartsNode node) throws XMLException{
			out.writeStartElement(PARTS_NODE);
			out.writeAttribute(TITLE, node.getTitle());
			
			if(node.hasLayout()){
				out.writeStartElement(PARTS);
				ElementWriter writer = getGraphWriter(node.getViewLayout(), null);
				String s = new FlowDataWriter().writeToString(new ElementWriter[]{writer});
				out.writeCharacters(s);
				out.writeEndElement();
			}
			
			if(!node.isLeaf()){
				for(int i = 0; i < node.getChildCount(); i++){
					PartsNode child = (PartsNode)node.getChildAt(i);
					writePartsNode(out, child);
				}
			}
			
			out.writeEndElement();
		}
	}
}
